@@ -2,10 +2,11 @@
  * ctcp.c:handles the client-to-client protocol(ctcp). 
  *
  * Written By Michael Sandrof 
- * Copyright(c) 1990, 1995 Michael Sandroff and others
+ * Copyright 1990, 1995 Michael Sandroff and others
  * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
  *
  * Serious cleanup by jfn (August 1996)
+ * Copyright 1996 EPIC Software Labs
  */
 
 #if 0
@@ -13,22 +14,10 @@ static	char	rcsid[] = "@(#)$Id: ctcp.c,v 1.16 1994/07/23 15:07:27 mrg Exp $";
 #endif
 
 #include "irc.h"
-#include "irc_std.h"
-
-#include <pwd.h>
-
-#ifdef HAVE_UNAME
-# include <sys/utsname.h>
-#endif
-
-#ifdef HAVE_STDARG_H
-#include <stdarg.h>
-#endif
-
 #include "crypt.h"
 #include "ctcp.h"
 #include "dcc.h"
-#include "edit.h"
+#include "commands.h"
 #include "hook.h"
 #include "ignore.h"
 #include "ircaux.h"
@@ -41,6 +30,12 @@ static	char	rcsid[] = "@(#)$Id: ctcp.c,v 1.16 1994/07/23 15:07:27 mrg Exp $";
 #include "vars.h"
 #include "window.h"
 #include "if.h"
+#include "flood.h"
+
+#include <pwd.h>
+#ifdef HAVE_UNAME
+# include <sys/utsname.h>
+#endif
 
 #define CTCP_SPECIAL	0	/* Limited, quiet, noreply, not expanded */
 #define CTCP_REPLY	1	/* Sends a reply (not significant) */
@@ -48,7 +43,7 @@ static	char	rcsid[] = "@(#)$Id: ctcp.c,v 1.16 1994/07/23 15:07:27 mrg Exp $";
 #define CTCP_NOLIMIT	4	/* Limit of one per privmsg. */
 #define CTCP_TELLUSER	8	/* Tell the user about it. */
 
-static	void 	split_CTCP _((char *, char *, char *));
+static	int 	split_CTCP (char *, char *, char *);
 
 /*
  * ctcp_entry: the format for each ctcp function.   note that the function
@@ -60,7 +55,7 @@ static	void 	split_CTCP _((char *, char *, char *));
  */
 
 struct _CtcpEntry;
-typedef char *(*CTCP_Handler) _((struct _CtcpEntry *, char *, char *, char *));
+typedef char *(*CTCP_Handler) (struct _CtcpEntry *, char *, char *, char *);
 typedef	struct _CtcpEntry
 {
 	char		*name;  /* name of ctcp datatag */
@@ -72,19 +67,19 @@ typedef	struct _CtcpEntry
 }	CtcpEntry;
 
 /* forward declarations for the built in CTCP functions */
-static	char	*do_sed 	_((CtcpEntry *, char *, char *, char *));
-static	char	*do_version 	_((CtcpEntry *, char *, char *, char *));
-static	char	*do_clientinfo 	_((CtcpEntry *, char *, char *, char *));
-static	char	*do_ping 	_((CtcpEntry *, char *, char *, char *));
-static	char	*do_echo 	_((CtcpEntry *, char *, char *, char *));
-static	char	*do_userinfo 	_((CtcpEntry *, char *, char *, char *));
-static	char	*do_finger 	_((CtcpEntry *, char *, char *, char *));
-static	char	*do_time 	_((CtcpEntry *, char *, char *, char *));
-static	char	*do_atmosphere 	_((CtcpEntry *, char *, char *, char *));
-static	char	*do_dcc 	_((CtcpEntry *, char *, char *, char *));
-static	char	*do_utc 	_((CtcpEntry *, char *, char *, char *));
-static	char	*do_dcc_reply 	_((CtcpEntry *, char *, char *, char *));
-static	char	*do_ping_reply 	_((CtcpEntry *, char *, char *, char *));
+static	char	*do_sed 	(CtcpEntry *, char *, char *, char *);
+static	char	*do_version 	(CtcpEntry *, char *, char *, char *);
+static	char	*do_clientinfo 	(CtcpEntry *, char *, char *, char *);
+static	char	*do_ping 	(CtcpEntry *, char *, char *, char *);
+static	char	*do_echo 	(CtcpEntry *, char *, char *, char *);
+static	char	*do_userinfo 	(CtcpEntry *, char *, char *, char *);
+static	char	*do_finger 	(CtcpEntry *, char *, char *, char *);
+static	char	*do_time 	(CtcpEntry *, char *, char *, char *);
+static	char	*do_atmosphere 	(CtcpEntry *, char *, char *, char *);
+static	char	*do_dcc 	(CtcpEntry *, char *, char *, char *);
+static	char	*do_utc 	(CtcpEntry *, char *, char *, char *);
+static	char	*do_dcc_reply 	(CtcpEntry *, char *, char *, char *);
+static	char	*do_ping_reply 	(CtcpEntry *, char *, char *, char *);
 
 
 static CtcpEntry ctcp_cmd[] =
@@ -95,10 +90,10 @@ static CtcpEntry ctcp_cmd[] =
 	{ "UTC",	CTCP_UTC, 	CTCP_INLINE | CTCP_NOLIMIT,
 		"substitutes the local timezone",
 		do_utc, 	do_utc },
-	{ "ACTION",	CTCP_ACTION, 	CTCP_SPECIAL,
+	{ "ACTION",	CTCP_ACTION, 	CTCP_SPECIAL | CTCP_NOLIMIT,
 		"contains action descriptions for atmosphere",
-		do_atmosphere, 	NULL },
-	{ "DCC",	CTCP_DCC, 	CTCP_SPECIAL,
+		do_atmosphere, 	do_atmosphere },
+	{ "DCC",	CTCP_DCC, 	CTCP_SPECIAL | CTCP_NOLIMIT,
 		"requests a direct_client_connection",
 		do_dcc, 	do_dcc_reply },
 	{ "VERSION",	CTCP_VERSION,	CTCP_REPLY | CTCP_TELLUSER,
@@ -146,15 +141,8 @@ int     sed = 0;
  */
 int	in_ctcp_flag = 0;
 
-#ifdef __STDC__
 #define CTCP_HANDLER(x) \
 	static char * x (CtcpEntry *ctcp, char *from, char *to, char *cmd)
-#else
-#define CTCP_HANDLER(x) 			\
-	static char *x (ctcp, from, to, cmd)	\
-	CtcpEntry   *ctcp;			\
-	char	    *from, *to, *cmd;
-#endif
 
 
 
@@ -172,10 +160,12 @@ CTCP_HANDLER(do_sed)
 		*crypt_who;
 	char	*ret = NULL, *ret2 = NULL;
 
-	if (my_stricmp(to, get_server_nickname(from_server)))
-		crypt_who = to;
-	else
+	if (*from == '=')		/* DCC CHAT message */
 		crypt_who = from;
+	else if (is_me(from_server, to))
+		crypt_who = from;
+	else
+		crypt_who = to;
 
 	if ((key = is_crypted(crypt_who)))
 		ret = decrypt_msg(cmd, key);
@@ -201,24 +191,25 @@ CTCP_HANDLER(do_utc)
 	if (!cmd || !*cmd)
 		return m_strdup(empty_string);
 
-	return m_strdup(my_ctime(atol(cmd)));
+	return m_strdup(my_ctime(my_atol(cmd)));
 }
 
 
 /*
  * do_atmosphere: does the CTCP ACTION command --- done by lynX
- * Changed this to make the default look less offensive to people
+ * Changed this to make the default look less offensive to users
  * who don't like it and added a /on ACTION. This is more in keeping
  * with the design philosophy behind IRCII
  */
 CTCP_HANDLER(do_atmosphere)
 {
-	int old;
+	const char	*old_message_from;
+	int	old_message_level;
 
 	if (!cmd || !*cmd)
 		return NULL;
 
-	old = set_lastlog_msg_level(LOG_ACTION);
+	save_message_from(&old_message_from, &old_message_level);
 	if (is_channel(to))
 	{
 		message_from(to, LOG_ACTION);
@@ -237,8 +228,7 @@ CTCP_HANDLER(do_atmosphere)
 			put_it("*> %s %s", from, cmd);
 	}
 
-	message_from(NULL, LOG_CRAP);
-	set_lastlog_msg_level(old);
+	restore_message_from(old_message_from, old_message_level);
 	return NULL;
 }
 
@@ -255,7 +245,7 @@ CTCP_HANDLER(do_dcc)
 	char	*size;
 	char	*extra_flags;
 
-	if (my_stricmp(to, get_server_nickname(from_server)))
+	if (!is_me(from_server, to))
 		return NULL;
 
 	if     (!(type = next_arg(cmd, &cmd)) ||
@@ -349,13 +339,15 @@ CTCP_HANDLER(do_version)
 	send_ctcp(CTCP_NOTICE, from, CTCP_VERSION, 
 			"ircII %s %s %s - %s", 
 			irc_version, the_unix, the_version,
+			(tmp = get_string_var(CLIENTINFO_VAR)) ? 
+				tmp : IRCII_COMMENT);
 #else
 	send_ctcp(CTCP_NOTICE, from, CTCP_VERSION, 
 			"ircII %s *IX - %s", 
 			irc_version,
+			(tmp = get_string_var(CLIENTINFO_VAR)) ? 
+				tmp : IRCII_COMMENT);
 #endif
-			(tmp = get_string_var(CLIENTINFO_VAR)) ? tmp : IRCII_COMMENT);
-
 	return NULL;
 }
 
@@ -370,8 +362,10 @@ CTCP_HANDLER(do_time)
 /* do_userinfo: does the CTCP USERINFO command */
 CTCP_HANDLER(do_userinfo)
 {
-	send_ctcp(CTCP_NOTICE, from, CTCP_USERINFO,
-			"%s", get_string_var(USERINFO_VAR));
+	char *tmp;
+
+	send_ctcp(CTCP_NOTICE, from, CTCP_USERINFO, "%s", 
+		(tmp = get_string_var(USERINFO_VAR)) ? tmp : "<No User Information>");
 	return NULL;
 }
 
@@ -382,7 +376,7 @@ CTCP_HANDLER(do_userinfo)
 CTCP_HANDLER(do_echo)
 {
 	if (!is_channel(to))
-		send_ctcp(CTCP_NOTICE, from, CTCP_ECHO, "%s", cmd);
+		send_ctcp(CTCP_NOTICE, from, ctcp->id, "%s", cmd);
 	return NULL;
 }
 
@@ -403,6 +397,13 @@ CTCP_HANDLER(do_finger)
 	char	*tmp;
 	char	*ctcpuser,
 		*ctcpfinger;
+	const char	*my_host;
+
+	if ((my_host = get_server_userhost(from_server)) &&
+			strchr(my_host, '@'))
+		my_host = strchr(my_host, '@') + 1;
+	else
+		my_host = hostname;
 
 	diff = time(NULL) - idle_time;
 
@@ -413,37 +414,19 @@ CTCP_HANDLER(do_finger)
 #define GECOS_DELIMITER ','
 #endif
 
-	if ((tmp = index(pwd->pw_gecos, GECOS_DELIMITER)) != NULL)
+	if ((tmp = strchr(pwd->pw_gecos, GECOS_DELIMITER)) != NULL)
 		*tmp = '\0';
 
-/* 
- * Three options for handling CTCP replies
- *  + Fascist Bastard Way -- normal, non-hackable fashion
- *  + Winterhawk way (default) -- allows hacking through IRCUSER and
- *	IRCFINGER environment variables
- *  + hop way -- returns a blank always
- */
-#if defined(HOP)
-	send_ctcp(CTCP_NOTICE, from, CTCP_FINGER, empty_string);
-#else
-# if !defined(I_DONT_TRUST_MY_USERS)
-	/* 
-	 * It would be pretty pointless to allow for customisable
-	 * usernames if they can track via IRCNAME from the
-	 * /etc/passwd file...  We therefore need to either disable
-	 * CTCP_FINGER or also make it customisable.  Let's do the
-	 * latter because it invokes less suspicion in the long run
-	 *				-- Jake [WinterHawk] Khuon
-	 */
+#if !defined(I_DONT_TRUST_MY_USERS)
 	if ((ctcpuser = getenv("IRCUSER"))) 
 		strmcpy(pwd->pw_name, ctcpuser, NAME_LEN);
 	if ((ctcpfinger = getenv("IRCFINGER"))) 
 		strmcpy(pwd->pw_gecos, ctcpfinger, NAME_LEN);
-# endif
+#endif
 	send_ctcp(CTCP_NOTICE, from, CTCP_FINGER, 
 		"%s (%s@%s) Idle %ld second%s", 
-		pwd->pw_gecos, pwd->pw_name, hostname, diff, plural(diff));
-#endif
+		pwd->pw_gecos, pwd->pw_name, my_host, diff, plural(diff));
+
 	return NULL;
 }
 
@@ -466,7 +449,7 @@ CTCP_HANDLER(do_dcc_reply)
 		type = next_arg(cmd, &cmd);
 
 	if (subcmd && type && !strcmp(subcmd, "REJECT"))
-		dcc_reject (from, type, cmd);
+		dcc_reject(from, type, cmd);
 
 	return NULL;
 }
@@ -477,23 +460,31 @@ CTCP_HANDLER(do_dcc_reply)
  */
 CTCP_HANDLER(do_ping_reply)
 {
-	char *sec, *usec = NULL;
-	struct timeval t;
-	time_t tsec = 0, tusec = 0;
+	struct timeval 	t;
+	time_t 		tsec = 0, 
+			tusec = 0, 
+			orig;
+	char *		ptr;
 
 	if (!cmd || !*cmd)
 		return NULL;		/* This is a fake -- cant happen. */
 
-	gettimeofday(&t, NULL);
+	orig = my_atol(cmd);
+	get_time(&t);
 
-	/* We've already checked 'cmd' here, so its safe. */
-	sec = cmd;
-	tsec = t.tv_sec - atoi(sec);
-
-	if ((usec = index(sec, ' ')))
+	/* Reply must be between time we started and right now */
+	if (orig < start_time || orig > t.tv_sec)
 	{
-		*usec++ = 0;
-		tusec = t.tv_usec - atoi(usec);
+		say("Invalid CTCP PING reply [%s] dropped.", cmd);
+		return NULL;
+	}
+
+	tsec = t.tv_sec - orig;
+
+	if ((ptr = strchr(cmd, ' ')))
+	{
+		*ptr++ = 0;
+		tusec = t.tv_usec - my_atol(ptr);
 	}
 
 	/*
@@ -528,19 +519,15 @@ CTCP_HANDLER(do_ping_reply)
  * scenario.  This is horrendously inefficient, since most privmsgs dont
  * contain any CTCPs.  So we check to see if there are any CTCPs in the
  * message before we bother doing anything.  THIS IS AN INELEGANT HACK!
- * But the call to word_count() is less expensive than even one copy to 
- * strcpy() since they both evaluate *each* character, and word_count()
+ * But the call to charcount() is less expensive than even one copy to 
+ * strcpy() since they both evaluate *each* character, and charcount()
  * doesnt have to do a write unless the character is present.  So it is 
  * definitely worth the cost to save CPU time for 99% of the PRIVMSGs.
  */
-#ifdef __STDC__
-extern 	char *do_ctcp (char *from, char *to, char *str)
-#else
-extern	char *do_ctcp (from, to, str)
-	char *from, *to, *str;
-#endif
+char *	do_ctcp (char *from, char *to, char *str)
 {
 	int 	flag;
+	int	fflag;
 	int	lastlog_level;
 	char 	local_ctcp_buffer [BIG_BUFFER_SIZE + 1],
 		the_ctcp          [IRCD_BUFFER_SIZE + 1],
@@ -560,16 +547,20 @@ static	time_t	last_ctcp_parsed = 0;
 		allow_ctcp_reply = 0;	/* Historical limit of 4 CTCPs */
 
 	flag = check_ignore_channel(from, FromUserHost, to, IGNORE_CTCPS);
-	if (!in_ctcp_flag)
-		in_ctcp_flag = 1;
-	strmcpy(local_ctcp_buffer, str, IRCD_BUFFER_SIZE);
+	fflag = new_check_flooding(from, is_channel(to) ? to : NULL,
+						str, CTCP_FLOOD); 
 
-	for (;;strmcat(local_ctcp_buffer, last, BIG_BUFFER_SIZE))
+	in_ctcp_flag++;
+	strmcpy(local_ctcp_buffer, str, IRCD_BUFFER_SIZE - 2);
+
+	lastlog_level = set_lastlog_msg_level(LOG_CTCP);
+	for (;;strmcat(local_ctcp_buffer, last, BIG_BUFFER_SIZE - 2))
 	{
-		split_CTCP(local_ctcp_buffer, the_ctcp, last);
+		if (split_CTCP(local_ctcp_buffer, the_ctcp, last))
+			break;		/* All done! */
 
 		if (!*the_ctcp)
-			break;		/* all done! */
+			continue;	/* Empty requests are ignored */
 
 		/*
 		 * Apply some integrety rules:
@@ -594,12 +585,12 @@ static	time_t	last_ctcp_parsed = 0;
 
 		/*
 		 * Check to see if the user is ignoring person.
+		 * Or if we're suppressing a flood.
 		 */
-		if (flag == IGNORED)
+		if (flag == IGNORED || fflag == 0)
 		{
-			if (x_debug)
+			if (x_debug & DEBUG_CTCPS)
 				yell("CTCP from [%s] ignored", from);
-
 			allow_ctcp_reply = 0;
 			continue;
 		}
@@ -609,24 +600,24 @@ static	time_t	last_ctcp_parsed = 0;
 		 */
 		if (get_int_var(NO_CTCP_FLOOD_VAR))
 		{
-			if (time(NULL) - last_ctcp_parsed < 2)
-			{
-				/*
-				 * This extends the flood protection until
-				 * we dont get a CTCP for 2 seconds.
-				 */
-				last_ctcp_parsed = time(NULL);
-				allow_ctcp_reply = 0;
-				if (x_debug)
-					say("CTCP flood from [%s] ignored", from);
-				continue;
-			}
+		    if (time(NULL) - last_ctcp_parsed < 2)
+		    {
+			/*
+			 * This extends the flood protection until
+			 * we dont get a CTCP for 2 seconds.
+			 */
+			last_ctcp_parsed = time(NULL);
+			allow_ctcp_reply = 0;
+			if (x_debug & DEBUG_CTCPS)
+				say("CTCP flood from [%s] ignored", from);
+			continue;
+		    }
 		}
 
 		/*
 		 * Check for global message
 		 */
-		if (*to == '$' || (*to == '#' && !lookup_channel(to, from_server, 0)))
+		if (*to == '$' || (*to == '#' && !im_on_channel(to, from_server)))
 		{
 			allow_ctcp_reply = 0;
 			continue;
@@ -638,12 +629,17 @@ static	time_t	last_ctcp_parsed = 0;
 		 * First we remove the argument.
 		 */
 		ctcp_command = the_ctcp;
-		ctcp_argument = index(the_ctcp, ' ');
+		ctcp_argument = strchr(the_ctcp, ' ');
 		if (ctcp_argument)
 			*ctcp_argument++ = 0;
 		else
 			ctcp_argument = empty_string;
 
+		/* Set up the window level/logging */
+		if (im_on_channel(to, from_server))
+			message_from(to, LOG_CTCP);
+		else
+			message_from(from, LOG_CTCP);
 
 		/*
 		 * Then we look for the correct CTCP.
@@ -661,12 +657,14 @@ static	time_t	last_ctcp_parsed = 0;
 			 * Offer it to the user.
 			 * Maybe they know what to do with it.
 			 */
-			if (do_hook(CTCP_LIST, "%s %s %s %s", from, to, ctcp_command, ctcp_argument))
+			if (do_hook(CTCP_LIST, "%s %s %s %s", from, to, 
+						ctcp_command, ctcp_argument))
 			{
 				if (get_int_var(VERBOSE_CTCP_VAR))
-					say("Unknown CTCP %s from %s to %s: %s%s",
-						ctcp_command, from, to, *ctcp_argument ? ": " :
-						empty_string, ctcp_argument);
+				    say("Unknown CTCP %s from %s to %s: %s%s",
+					ctcp_command, from, to, 
+					*ctcp_argument ? ": " : empty_string, 
+					ctcp_argument);
 			}
 			time(&last_ctcp_parsed);
 			allow_ctcp_reply = 0;
@@ -676,7 +674,13 @@ static	time_t	last_ctcp_parsed = 0;
 		/* 
 		 * We did find it.  Acknowledge it.
 		 */
-		ptr = ctcp_cmd[i].func(ctcp_cmd + i, from, to, ctcp_argument);
+		ptr = empty_string;
+		if (do_hook(CTCP_REQUEST_LIST, "%s %s %s %s",
+				from, to, ctcp_command, ctcp_argument))
+		{
+			ptr = ctcp_cmd[i].func(ctcp_cmd + i, from, 
+						to, ctcp_argument);
+		}
 
 		/*
 		 * If this isnt an 'unlimited' CTCP, set up flood protection.
@@ -710,35 +714,39 @@ static	time_t	last_ctcp_parsed = 0;
 		 */
 		if (ctcp_cmd[i].flag & CTCP_TELLUSER)
 		{
-			if (do_hook(CTCP_LIST, "%s %s %s %s", from, to, ctcp_command, ctcp_argument))
+		    if (do_hook(CTCP_LIST, "%s %s %s %s", 
+				from, to, ctcp_command, ctcp_argument))
+		    {
+			if (get_int_var(VERBOSE_CTCP_VAR)) 
 			{
-				/* Set up the window level/logging */
-				lastlog_level = set_lastlog_msg_level(LOG_CTCP);
-				message_from(NULL, LOG_CTCP);
-
-				if (get_int_var(VERBOSE_CTCP_VAR))
-				if (!my_stricmp(to, get_server_nickname(from_server)))
-					say("CTCP %s from %s%s%s", 
-						ctcp_command, from, 
-						ctcp_argument && *ctcp_argument ? ": " : empty_string, 
-						ctcp_argument && *ctcp_argument ? ctcp_argument : empty_string);
-				else
-					say("CTCP %s from %s to %s%s%s",
-						ctcp_command, from, to, 
-						ctcp_argument && *ctcp_argument ? ": " : empty_string, 
-						ctcp_argument && *ctcp_argument ? ctcp_argument : empty_string);
-
-				/* Reset the window level/logging */
-				set_lastlog_msg_level(lastlog_level);
+			    if (is_me(from_server, to))
+				say("CTCP %s from %s%s%s", 
+					ctcp_command, from, 
+					*ctcp_argument ? ": " : empty_string, 
+					ctcp_argument);
+			    else
+				say("CTCP %s from %s to %s%s%s",
+					ctcp_command, from, to, 
+					*ctcp_argument ? ": " : empty_string, 
+					ctcp_argument);
 			}
+		    }
 		}
-
-		new_free(&ptr);
+		if (ptr != empty_string)
+			new_free(&ptr);
 	}
 
-	if (in_ctcp_flag == 1)
-		in_ctcp_flag = 0;
+	/* Reset the window level/logging */
+	set_lastlog_msg_level(lastlog_level);
+	message_from(NULL, LOG_CRAP);
 
+	in_ctcp_flag--;
+
+	/* 
+	 * Since 'local_ctcp_buffer' is derived from 'str',
+	 * strlen(local_ctcp_buffer) <= strlen(str) in all circumstances,
+	 * so this strcpy is safe.
+	 */
 	return strcpy(str, local_ctcp_buffer);
 }
 
@@ -748,12 +756,7 @@ static	time_t	last_ctcp_parsed = 0;
  * do_notice_ctcp: a re-entrant form of a CTCP reply parser.
  * See the implementation notes in do_ctcp().
  */
-#ifdef __STDC__
-extern 	char *do_notice_ctcp (char *from, char *to, char *str)
-#else
-extern	char *do_notice_ctcp (from, to, str)
-	char *from, *to, *str;
-#endif
+char *	do_notice_ctcp (char *from, char *to, char *str)
 {
 	int 	flag;
 	int	lastlog_level;
@@ -776,13 +779,15 @@ extern	char *do_notice_ctcp (from, to, str)
 	flag = check_ignore_channel(from, FromUserHost, to, IGNORE_CTCPS);
 	if (!in_ctcp_flag)
 		in_ctcp_flag = -1;
-	strmcpy(local_ctcp_buffer, str, IRCD_BUFFER_SIZE);
+	strmcpy(local_ctcp_buffer, str, IRCD_BUFFER_SIZE - 2);
 
 	for (;;strmcat(local_ctcp_buffer, last, BIG_BUFFER_SIZE))
 	{
-		split_CTCP(local_ctcp_buffer, the_ctcp, last);
+		if (split_CTCP(local_ctcp_buffer, the_ctcp, last))
+			break;		/* All done! */
+
 		if (!*the_ctcp)
-			break;		/* all done! */
+			continue;	/* Empty requests are ignored */
 
 		/*
 		 * Apply sanity rules
@@ -793,14 +798,14 @@ extern	char *do_notice_ctcp (from, to, str)
 
 		if (flag == IGNORED)
 		{
-			if (x_debug)
+			if (x_debug & DEBUG_CTCPS)
 				yell("CTCP REPLY from [%s] ignored", from);
 			allow_ctcp_reply = 0;
 			continue;
 		}
 
 		/* Global messages -- just drop the CTCP */
-		if (*to == '$' || (*to == '#' && !lookup_channel(to, from_server, 0)))
+		if (*to == '$' || (*to == '#' && !im_on_channel(to, from_server)))
 		{
 			allow_ctcp_reply = 0;
 			continue;
@@ -812,7 +817,7 @@ extern	char *do_notice_ctcp (from, to, str)
 		 */
 
 		ctcp_command = the_ctcp;
-		ctcp_argument = index(the_ctcp, ' ');
+		ctcp_argument = strchr(the_ctcp, ' ');
 		if (ctcp_argument)
 			*ctcp_argument++ = 0;
 		else
@@ -861,13 +866,17 @@ extern	char *do_notice_ctcp (from, to, str)
 	if (in_ctcp_flag == -1)
 		in_ctcp_flag = 0;
 
+	/* 
+	 * local_ctcp_buffer is derived from 'str', so its always
+	 * smaller or equal in size to 'str', so this copy is safe.
+	 */
 	return strcpy(str, local_ctcp_buffer);
 }
 
 
 
 /* in_ctcp: simply returns the value of the ctcp flag */
-extern int	in_ctcp _((void)) { return (in_ctcp_flag); }
+int	in_ctcp (void) { return (in_ctcp_flag); }
 
 
 
@@ -881,38 +890,45 @@ extern int	in_ctcp _((void)) { return (in_ctcp_flag); }
  * transparantly.  This greatly reduces the logic, complexity, and
  * possibility for error in this function.
  */
-#if defined(__STDC__) && defined(HAVE_STDARG_H)
-extern	void	send_ctcp(int type, char *to, int datatag, char *format, ...)
-#else
-extern	void	send_ctcp(type, to, datatag, format, arg0, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9)
-	int	type, datatag;
-	char 	*to, *format,
-     		*arg0, *arg1, *arg2, *arg3, *arg4, 
-		*arg5, *arg6, *arg7, *arg8, *arg9;
-#endif
+void	send_ctcp (int type, char *to, int datatag, char *format, ...)
 {
-	char putbuf [BIG_BUFFER_SIZE + 1],
-	     putbuf2[BIG_BUFFER_SIZE + 1];
+	char 	putbuf [BIG_BUFFER_SIZE + 1],
+		*putbuf2;
+	int	len;
 
-	if (in_on_who)
-		return;
+	/* Make sure that the final \001 doesnt get truncated */
+	len = IRCD_BUFFER_SIZE - (12 + strlen(to));
+	putbuf2 = alloca(len);
 
 	if (format)
 	{
-#if defined(__STDC__) && defined(HAVE_STDARG_H)
 		va_list args;
 		va_start(args, format);
-		vsprintf(putbuf, format, args);
+		vsnprintf(putbuf, BIG_BUFFER_SIZE, format, args);
 		va_end(args);
-#else
-		sprintf(putbuf, format, arg0, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9);
-#endif
 
-		sprintf(putbuf2, "%c%s %s%c", CTCP_DELIM_CHAR, ctcp_cmd[datatag].name, putbuf, CTCP_DELIM_CHAR);
+		do_hook(SEND_CTCP_LIST, "%s %s %s %s", 
+				ctcp_type[type], to, 
+				ctcp_cmd[datatag].name, putbuf);
+		snprintf(putbuf2, len, "%c%s %s%c", 
+				CTCP_DELIM_CHAR, 
+				ctcp_cmd[datatag].name, putbuf, 
+				CTCP_DELIM_CHAR);
 	}
 	else
-		sprintf(putbuf2, "%c%s%c", CTCP_DELIM_CHAR, ctcp_cmd[datatag].name, CTCP_DELIM_CHAR);
+	{
+		do_hook(SEND_CTCP_LIST, "%s %s %s", 
+				ctcp_type[type], to, 
+				ctcp_cmd[datatag].name);
+		snprintf(putbuf2, len, "%c%s%c", 
+				CTCP_DELIM_CHAR, 
+				ctcp_cmd[datatag].name, 
+				CTCP_DELIM_CHAR);
+	}
 
+	/* XXX - Ugh.  What a hack. */
+	putbuf2[len - 2] = CTCP_DELIM_CHAR;
+	putbuf2[len - 1] = 0;
 
 	send_text(to, putbuf2, ctcp_type[type], 0);
 }
@@ -924,13 +940,7 @@ extern	void	send_ctcp(type, to, datatag, format, arg0, arg1, arg2, arg3, arg4, a
  * null terminated (it can contain nulls).  Returned is a malloced, null
  * terminated string.   
  */
-#ifdef __STDC__
-extern 	char	*ctcp_quote_it (char *str, int len)
-#else
-extern 	char	*ctcp_quote_it(str, len)
-	char	*str;
-	int	len;
-#endif
+char	*ctcp_quote_it (char *str, int len)
 {
 	char	buffer[BIG_BUFFER_SIZE + 1];
 	char	*ptr;
@@ -971,13 +981,7 @@ extern 	char	*ctcp_quote_it(str, len)
  * convenied, but the returned data may contain nulls!.  The len is modified
  * to contain the size of the data returned. 
  */
-#ifdef __STDC__
-extern	char	*ctcp_unquote_it (char *str, int *len)
-#else
-extern	char	*ctcp_unquote_it(str, len)
-	char	*str;
-	int	*len;
-#endif
+char	*ctcp_unquote_it (char *str, int *len)
 {
 	char	*buffer;
 	char	*ptr;
@@ -985,7 +989,7 @@ extern	char	*ctcp_unquote_it(str, len)
 	int	i,
 		new_size = 0;
 
-	buffer = (char *) new_malloc(sizeof(char) * *len);
+	buffer = (char *) new_malloc(sizeof(char) * *len + 1);
 	ptr = buffer;
 	i = 0;
 	while (i < *len)
@@ -1023,12 +1027,7 @@ extern	char	*ctcp_unquote_it(str, len)
 	return (buffer);
 }
 
-#ifdef __STDC__
-extern int get_ctcp_val (char *str)
-#else
-extern int get_ctcp_val (str)
-char *str;
-#endif
+int 	get_ctcp_val (char *str)
 {
 	int i;
 
@@ -1054,26 +1053,25 @@ char *str;
  * XXXX -- some may call this a hack, but if youve got a better
  * way to handle this job, id love to use it.
  */
-void split_CTCP(char *raw_message, char *ctcp_dest, char *after_ctcp)
+static int split_CTCP (char *raw_message, char *ctcp_dest, char *after_ctcp)
 {
-	char *ctcp_start, *ctcp_end;
+	char 	*ctcp_start, 
+		*ctcp_end;
 
 	*ctcp_dest = *after_ctcp = 0;
-	ctcp_start = index(raw_message, CTCP_DELIM_CHAR);
-	if (!ctcp_start)
-		return;		/* No CTCPs present. */
 
+	if (!(ctcp_start = strchr(raw_message, CTCP_DELIM_CHAR)))
+		return -1;		/* No CTCPs present. */
 	*ctcp_start++ = 0;
-	ctcp_end = index(ctcp_start, CTCP_DELIM_CHAR);
-	if (!ctcp_end)
+
+	if (!(ctcp_end = strchr(ctcp_start, CTCP_DELIM_CHAR)))
 	{
-		*--ctcp_start = CTCP_DELIM_CHAR;
-		return;		/* Thats _not_ a CTCP. */
+		*--ctcp_start = CTCP_DELIM_CHAR; /* Revert change */
+		return -1;		 /* No CTCPs present after all */
 	}
-
 	*ctcp_end++ = 0;
-	strmcpy(ctcp_dest, ctcp_start, IRCD_BUFFER_SIZE);
-	strmcpy(after_ctcp, ctcp_end, IRCD_BUFFER_SIZE);
 
-	return;		/* All done! */
+	strmcpy(ctcp_dest, ctcp_start, IRCD_BUFFER_SIZE - 2);
+	strmcpy(after_ctcp, ctcp_end, IRCD_BUFFER_SIZE - 2);
+	return 0;		/* All done! */
 }
